/*
 * SPI boot code - inspired by the Cirrus EDB9307 dev kit code.
 *
 * Copyright (C) Ben Dubb, IPC Corporation.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */



#include <config.h>
#include <version.h>
#include <ep93xx.h>


.macro ToggleRedLed
        ldr     r1, =EP9312_LED_DATA
        ldr     r0, [r1]
        eor     r0, r0, #(EP9312_LED_RED_ON)
        str     r0, [r1]
.endm

.macro TurnOffRedLed
        ldr     r1, =EP9312_LED_DATA
        ldr     r0, [r1]
        bic     r0, r0, #(EP9312_LED_RED_ON)
        str     r0, [r1]
.endm

		
SPI_BASE_ADDR:	 .word 0x808a0000


				
///*****************************************************************************/
//
// The routine for reading a block of data from the SPI Flash
//   r0 - offset to start reading from
//   r1 - len
//   r2 - dst address
//
///*****************************************************************************/
		
.globl copy_spi
copy_spi:

    //
    // Save the offset to be read.
    //
    mov     r5, r1, lsr #2
    mov     r6, r2
    mov		r7, r0

    /* save the return address */
    mov ip, lr

    //
    // SDRAM had to been configured before we move code into it.
    //	
    nop
    nop
    nop
    nop
			
    //
    // Load a pointer to the SPI registers.
    //
    ldr     r0, SPI_BASE_ADDR
		
    //
    // Program the CR0 register.
    //
    ldr     r1, =0x000001c7
    str     r1, [r0, #0x00]

    //
    // Program the predivisor register.
    //
    mov     r1, #2
    str     r1, [r0, #0x10]

    //
    // Program the CR1 register.
    //
    mov     r1, #0
    str     r1, [r0, #0x04]

    //
    // Enable the SPI interface.
    //
    mov     r1, #0x10
    str     r1, [r0, #0x04]


read_word:	
    //
    // Load a pointer to the SPI registers.
    //
    ldr     r0, SPI_BASE_ADDR


    // load offset
    mov     r4, r7

    //
    // Send the SPI read command.
    //
    mov     r1, #0x03
    str     r1, [r0, #8]
    mov     r1, r4, lsr #16
    str     r1, [r0, #8]
    mov     r1, r4,	lsr #8
    str     r1, [r0, #8]
    str     r4, [r0, #8]
		
    //
    // Read a byte from the SPI to make sure that the FIFO isn't overrun.
    //
    bl      SPIReadByte

    //
    // The SPI can not read unless it is writing, so write out four zero bytes
    // so that four bytes can be read from the EEPROM.
    //
    mov     r1, #0x00
    str     r1, [r0, #0x08]
    str     r1, [r0, #0x08]
    str     r1, [r0, #0x08]
    str     r1, [r0, #0x08]

    //
    // Read and throw away the bytes corresponding to the sent address.
    //
    bl      SPIReadByte
    bl      SPIReadByte
    bl      SPIReadByte

		
    //
    // Read the data word from the EEPROM.
    //
    bl      SPIReadByte
    mov     r4, r1
    bl      SPIReadByte
    mov     r1, r1, lsl #8
    orr     r4, r4, r1
    bl      SPIReadByte
    mov	    r1, r1, lsl #16
    orr     r4, r4, r1
    bl      SPIReadByte
    mov     r1, r1, lsl #24
    orr     r4, r4, r1

    //
    // Get the SPI receive FIFO empty flag.
    //
    mov     r1, #0x04

    //
    // Read the SPI status register.
    //

receive_clear1:
    ldr     r2, [r0, #0x0c]

    //
    // Stop looping if there is no more data in the receive FIFO.
    //
    tst     r2, r1
    beq     receive_clear2

    //
    // Read a byte from the receive FIFO and loop back.
    //
    ldr     r2, [r0, #0x08]
    b       receive_clear1

receive_clear2:


	// Save the word
	str     r4, [r6]
		
	//
	// Loop until all (r5) words been copied
	//
	sub		r5, r5, #1
	add		r6, r6, #4
	add		r7, r7, #4
	cmp		r5, #0
	bne		read_word

	TurnOffRedLed
			
	/* restore the return address */
	mov lr, ip
		
    //
    // Return to the caller.
    //
    mov     r0, #1
	mov     pc, lr


				
///******************************************************************************/
//
// Reads a byte of data from the SPI port.
//
///******************************************************************************/

SPIReadByte:
    //
    // Get the flag to wait for (receive FIFO empty).
    //
    mov     r1, #0x04

    //
    // Load the SPI status register.
    //
spi_read1:
    ldr     r2, [r0, #0x0c]

    //
    // See if the bit is clear and loop back if so.
    //
    tst     r2, r1
    beq     spi_read1

    //
    // Read the byte from the SPI.
    //
    ldr     r1, [r0, #0x08]

    //
    // Return to the caller.
    //
    mov     pc, lr

